# import nats
# import typing as t
# from skoll.domain import Message
# from skoll.result import Result, ok
# from skoll.config import NATSConfig
# from skoll.application.types import RawMessage
# from skoll.application import Mediator, Subscriber, Service


# __all__ = ["NATSMediator"]


# class NATSMediator(Mediator):
#     """NATS mediator implementation."""

#     seed: str
#     urls: list[str]
#     subscribers: list[Subscriber]
#     handlers: dict[str, Subscriber]

#     def __init__(self, config: NATSConfig) -> None:
#         if config.urls is None or config.seed is None:
#             raise ValueError("NATS URLs and seed are required")
#         self.urls = config.urls.split(",")
#         self.seed = config.seed
#         self.handlers = {}
#         self.subscribers = []

#     @t.override
#     def register(self, *services: Service) -> None:
#         for service in services:
#             for subscriber in service.subscribers:
#                 if subscriber.with_reply:
#                     self.handlers[subscriber.topic] = subscriber
#                 else:
#                     self.subscribers.append(subscriber)

#     @t.override
#     async def start(self) -> None:
#         pass

#     @t.override
#     async def stop(self) -> None:
#         pass

#     @t.override
#     async def publish(self, msg: Message | RawMessage) -> None:
#         pass

#     @t.override
#     async def request(self, msg: Message | RawMessage) -> Result[t.Any]:
#         return ok(None)



# import typing as t
# from json import loads
# from attrs import define
# from re import match as re_match
# from aiohttp import ClientSession

# from skoll.utils import sanitize_dict
# from skoll.config import SpiceDBConfig
# from skoll.errors import Forbidden, InternalError
# from skoll.application import AuthzWriteChange, AuthzPrecondition, AuthzLookupResult, Authz


# __all__ = ["SpiceDBAuthz"]


# TUPLE_PATTERN = r"(?P<resource>.+?):(?P<resource_id>.*?)#(?P<relation>.*?)@(?P<subject>.*?):(?P<subject_id>.*)"


# @define(frozen=True, kw_only=True)
# class TupleObject:
#     resource: str
#     subject: str | None = None
#     relation: str | None = None
#     subject_id: str | None = None
#     resource_id: str | None = None
#     subject_relation: str | None = None


# class SpiceDBAuthz(Authz):

#     url: str
#     token: str

#     def __init__(self, config: SpiceDBConfig):
#         if config.uri is None or config.token is None:
#             raise InternalError(debug={"message": "Invalid SpiceDB config", "config": config})
#         self.url = config.uri
#         self.token = config.token

#     def make_headers(self) -> dict[str, str]:
#         return {"Authorization": f"Bearer {self.token}"}

#     @t.override
#     async def write(self, changes: list[AuthzWriteChange], preconditions: list[AuthzPrecondition] | None = None) -> str:
#         uri = self.url + "/v1/relationships/write"
#         data = {"updates": get_changes(changes), "optionalAuthzPreconditions": get_preconditions(preconditions or [])}
#         async with ClientSession() as session:
#             async with session.post(uri, json=sanitize_dict(data), ssl=False, headers=self.make_headers()) as response:
#                 if response.status != 200:
#                     print(await response.text())
#                     raise ValueError(f"Failed to write changes {changes} with preconditions {preconditions}")
#                 return (await response.json()).get("writtenAt", {}).get("token", "")

#     @t.override
#     async def lookup(
#         self, filter: str, cxt: dict[str, t.Any] | None = None, limit: int | None = None, cursor: str | None = None
#     ) -> AuthzLookupResult:
#         tuple_obj = tuple_from(filter)
#         if not tuple_obj:
#             raise ValueError(f"Invalid tuple: {filter}")

#         data: dict[str, t.Any] = {
#             "context": cxt,
#             "consistency": {"minimizeLatency": True},
#             "optionalCursor": {"token": cursor} if cursor else None,
#         }
#         uri = self.url + "/v1/permissions"
#         if tuple_obj.subject_id is not None:  # ResourceLookup
#             uri += "/resources"
#             data["optionalLimit"] = limit
#             data["permission"] = tuple_obj.relation
#             data["resourceObjectType"] = tuple_obj.resource
#             data["subject"] = {
#                 "optionalRelation": tuple_obj.subject_relation,
#                 "object": {"objectType": tuple_obj.subject, "objectId": tuple_obj.subject_id},
#             }
#         else:  # SubjectLookup
#             uri += "/subjects"
#             data["optionalConcreteLimit"] = limit
#             data["permission"] = tuple_obj.relation
#             data["subjectObjectType"] = tuple_obj.subject
#             data["optionalSubjectRelation"] = tuple_obj.subject_relation
#             data["resource"] = {"objectType": tuple_obj.resource, "objectId": tuple_obj.resource_id}

#         async with ClientSession() as session:
#             async with session.post(uri, json=sanitize_dict(data), ssl=False, headers=self.make_headers()) as response:
#                 if response.status != 200:
#                     raise Exception(
#                         f"Failed to lookup  {filter} with context {cxt} and limit {limit} and cursor {cursor}. status code: {response.status}"
#                     )
#                 lines = [loads(line) for line in (await response.text()).split("\n")[:-1]]
#                 id_key = "resourceObjectId" if tuple_obj.subject_id is not None else "subjectObjectId"
#                 return AuthzLookupResult(
#                     uids=[line.get("result", {}).get(id_key, "") for line in lines],
#                     cursor=lines[-1].get("result", {}).get("afterResultCursor", {}).get("token"),
#                 )

#     @t.override
#     async def check(self, tuple: str, cxt: dict[str, t.Any] | None = None) -> None:
#         uri = self.url + "/v1/permissions/check"
#         tuple_obj = tuple_from(tuple)
#         if not tuple_obj:
#             raise ValueError(f"Invalid tuple: {tuple}")

#         data: dict[str, t.Any] = {
#             "context": cxt,
#             "permission": tuple_obj.relation,
#             "consistency": {"minimizeLatency": True},
#             "resource": {"objectType": tuple_obj.resource, "objectId": tuple_obj.resource_id},
#             "subject": {
#                 "optionalRelation": tuple_obj.subject_relation,
#                 "object": {"objectType": tuple_obj.subject, "objectId": tuple_obj.subject_id},
#             },
#         }

#         async with ClientSession() as session:
#             async with session.post(uri, json=sanitize_dict(data), ssl=False, headers=self.make_headers()) as response:
#                 res = await response.json()
#                 if res.get("permissionship", "") != "PERMISSIONSHIP_HAS_PERMISSION":
#                     raise Forbidden(debug={"tuple": tuple, "context": cxt})
#                 return None


# def tuple_from(tuple_str: str) -> TupleObject | None:
#     match = re_match(TUPLE_PATTERN, tuple_str)
#     if match:
#         data = match.groupdict()
#         subject_id = data["subject_id"].split("#")
#         data["subject_id"] = subject_id[0]
#         data["subject_relation"] = subject_id[1] if len(subject_id) > 1 else None
#         return TupleObject(
#             resource=data["resource"],
#             subject_relation=data["subject_relation"],
#             subject=data["subject"] if len(data["subject"]) > 0 else None,
#             relation=data["relation"] if len(data["relation"]) > 0 else None,
#             subject_id=data["subject_id"] if len(data["subject_id"]) > 0 else None,
#             resource_id=data["resource_id"] if len(data["resource_id"]) > 0 else None,
#         )
#     return None


# def get_changes(changes: list[AuthzWriteChange]) -> list[dict[str, t.Any]]:
#     updates = []
#     for ch in changes:
#         tuple_obj = tuple_from(ch[1])
#         if not tuple_obj:
#             raise ValueError(f"Invalid tuple: {ch[1]}")
#         update: dict[str, t.Any] = {
#             "operation": "OPERATION_DELETE" if ch[0] == "DELETE" else "OPERATION_TOUCH",
#             "relationship": {
#                 "relation": tuple_obj.relation,
#                 "resource": {"objectType": tuple_obj.resource, "objectId": tuple_obj.resource_id},
#                 "subject": {
#                     "optionalRelation": tuple_obj.subject_relation,
#                     "object": {"objectType": tuple_obj.subject, "objectId": tuple_obj.subject_id},
#                 },
#                 "optionalExpiresAt": ch[3].iso_format if ch[3] else None,
#                 "optionalCaveat": {"caveatName": ch[2][0], "context": ch[2][1]} if ch[2] else None,
#             },
#         }
#         updates.append(update)
#     return updates


# def get_preconditions(precondition: list[AuthzPrecondition]) -> list[dict[str, t.Any]]:
#     return [{"operation": f"OPERATION_{pr[0]}", "filter": relation_filter_from(pr[1])} for pr in precondition]


# def relation_filter_from(tuple: str) -> dict[str, t.Any]:
#     tuple_obj = tuple_from(tuple)
#     if not tuple_obj:
#         raise ValueError(f"Invalid tuple: {tuple}")
#     return {
#         "resourceType": tuple_obj.resource,
#         "optionalResourceId": tuple_obj.resource_id,
#         "optionalResourceIdPrefix": tuple_obj.resource_id,
#         "optionalRelation": tuple_obj.relation,
#         "optionalSubjectFilter": {
#             "subjectType": tuple_obj.subject,
#             "optionalSubjectId": tuple_obj.subject_id,
#             "optionalRelation": {"relation": tuple_obj.subject_relation},
#         },
#     }



# import typing as t

# from skoll.domain import Message
# from skoll.result import Result, fail, is_fail
# from skoll.utils import call_with_dependencies
# from skoll.application import Subscriber, RawMessage
# from skoll.errors import Error, InternalError, ValidationFailed


# async def run_callback(subscriber: Subscriber, message: Message | RawMessage) -> Result[t.Any]:
#     topic = message.topic if isinstance(message, Message) else message.get("topic")
#     try:
#         msg = get_message(subscriber.msg_cls, message)
#         return await call_with_dependencies(subscriber.callback, {subscriber.msg_arg: msg})
#     except Error as err:
#         return fail(err=err)
#     except Exception as exc:
#         return fail(err=InternalError.from_exception(exc, extra={"subject": topic, "message": message}))


# def get_message(cls: type[Message], message: Message | RawMessage) -> Message:
#     if isinstance(message, Message):
#         return message
#     res = cls.create(raw=message)
#     if is_fail(res):
#         raise ValidationFailed(errors=res.err.errors)

#     return res.value


# def is_subscribed(subscriber: Subscriber, message: Message | RawMessage) -> bool:
#     topic = message.topic if isinstance(message, Message) else message.get("topic")
#     s = topic.split(".") if topic else []
#     p = subscriber.topic.split(".") if subscriber.topic else []

#     i = j = 0
#     while i < len(s) and j < len(p):
#         if j + 1 < len(p) and p[j + 1] == ">":
#             # Next token is '>', so this token must match exactly
#             if p[j] != "*" and p[j] != s[i]:
#                 return False
#             j += 2
#             i += 1
#             continue

#         if p[j] == ">":
#             return True

#         if p[j] == "*":
#             i += 1
#             j += 1
#             continue

#         if p[j] != s[i]:
#             return False

#         i += 1
#         j += 1

#     if i == len(s):
#         # Check remaining pattern tokens
#         while j < len(p):
#             if p[j] == ">":
#                 return True
#             if p[j] != "*":
#                 return False
#             j += 1
#         return True

#     # Subject has leftover tokens
#     return j < len(p) and p[j] == ">"



import typing as t
from attrs import define, field
from abc import ABC, abstractmethod

from skoll.domain import ID


__all__ = ["Criteria", "ListPage", "ListCriteria"]


class ListPage[T](t.NamedTuple):

    items: list[T]
    cursor: str | None = None


class SQLCriteria(t.NamedTuple):
    query: str
    params: list[t.Any]


@define(frozen=True, kw_only=True, slots=True)
class Criteria(ABC):

    uid: ID | None = None

    @property
    @abstractmethod
    def as_sql(self) -> SQLCriteria:
        raise NotImplementedError("Subclasses must implement this method")


@define(frozen=True, kw_only=True, slots=True)
class ListCriteria(Criteria, ABC):

    cursor: str | None = None
    limit: int = field(default=100)


# type AuthzWriteOperation = t.Literal["SET", "DELETE"]
# type AuthzPrecondition = tuple[AuthzPreconditionOperation, str]
# type AuthzPreconditionOperation = t.Literal["MUST_MATCH", "MUST_NOT_MATCH"]
# type SubscriberCallback = t.Callable[t.Concatenate[Message, ...], c.Coroutine[t.Any, t.Any, Result[t.Any]]]
# type AuthzWriteChange = tuple[AuthzWriteOperation, str, tuple[str, dict[str, t.Any]] | None, DateTime | None]


# class AuthzLookupResult(t.NamedTuple):
#     uids: list[str]
#     cursor: str | None = None

# class RawMessage(t.TypedDict):
#     topic: str
#     source: str
#     id: t.NotRequired[str]
#     created_at: t.NotRequired[int]
#     payload: t.NotRequired[dict[str, t.Any]]
#     context: t.NotRequired[dict[str, t.Any]]


# @define(frozen=True, kw_only=True, slots=True)
# class Subscriber:

#     topic: str
#     msg_arg: str
#     service_name: str
#     msg_cls: type[Message]
#     with_reply: bool = True
#     callback: SubscriberCallback


# @define(kw_only=True, slots=True)
# class MissingSubscriber(NotFound):

#     attr: str | None = field(default=None, init=False)
#     code: str = field(default="missing_subscriber", init=False)
#     detail: str = "No subscriber found for the given message subject"


# @define(kw_only=True, slots=True, frozen=True)
# class Service:

#     name: str
#     subscribers: list[Subscriber] = field(factory=list)

#     def on(self, topic: str, with_reply: bool = True):
#         def decorator(callback: SubscriberCallback):
#             first_arg = list(signature(callback).parameters.values())[0]
#             if not issubclass(first_arg.annotation, Message):
#                 raise TypeError(
#                     f"@service.on only allow on function with first argument being a subclass of Message, got {first_arg.annotation}"
#                 )

#             self.subscribers.append(
#                 Subscriber(
#                     topic=topic,
#                     callback=callback,
#                     with_reply=with_reply,
#                     msg_arg=first_arg.name,
#                     service_name=self.name,
#                     msg_cls=first_arg.annotation,
#                 )
#             )
#             return callback

#         return decorator



import typing as t
import collections.abc as c

from skoll.application.types import *
from skoll.domain import EntityState


__all__ = ["DB", "Repository"]


# class Broker(t.Protocol):

#     async def stop(self) -> None: ...
#     async def publish(self, msg: Message | RawMessage) -> None: ...
#     async def request(self, msg: Message | RawMessage) -> Result[t.Any]: ...
#     async def connect(self, *subscribers: list[Subscriber]) -> None: ...


# class Authz(t.Protocol):

#     async def check(self, tuple: str, cxt: dict[str, t.Any] | None = None) -> None: ...
#     async def write(
#         self, changes: list[AuthzWriteChange], preconditions: list[AuthzPrecondition] | None = None
#     ) -> str: ...
#     async def lookup(
#         self, filter: str, cxt: dict[str, t.Any] | None = None, limit: int | None = None, cursor: str | None = None
#     ) -> AuthzLookupResult: ...


class Repository[T: EntityState](t.Protocol):

    async def save(self, state: T) -> None: ...
    async def delete(self, criteria: Criteria) -> None: ...
    async def exist(self, criteria: Criteria) -> bool: ...
    async def get(self, criteria: Criteria) -> T | None: ...
    async def list(self, criteria: ListCriteria) -> ListPage[T]: ...


class DB[T = t.Any](t.Protocol):

    async def close(self) -> None: ...
    async def connect(self) -> None: ...
    def session(self) -> c.AsyncGenerator[T]: ...
    def transaction(self) -> c.AsyncGenerator[T]: ...



# import typing as t
# from attrs import define, field
# from asyncio import Queue, Event, create_task, wait

# from skoll.domain import Message
# from skoll.result import Result, fail
# from skoll.application import Subscriber, MissingSubscriber, Mediator, Service, RawMessage

# from .utils import run_callback, is_subscribed

# __all__ = ["BasicMediator"]


# @define(kw_only=True, slots=True)
# class BasicMediator(Mediator):
#     """Basic in-memory mediator implementation."""

#     queue: Queue[Message | RawMessage] = field(factory=Queue)
#     stop_signal: Event = field(factory=Event)
#     subscribers: list[Subscriber] = field(factory=list)
#     handlers: dict[str, Subscriber] = field(factory=dict)

#     @t.override
#     async def publish(self, msg: Message | RawMessage) -> None:
#         self.queue.put_nowait(msg)

#     @t.override
#     async def disconnect(self):
#         self.stop_signal.set()

#     @t.override
#     async def request(self, msg: Message | RawMessage) -> Result[t.Any]:
#         topic = msg.topic if isinstance(msg, Message) else msg["topic"]
#         subscriber = self.handlers.get(topic)
#         if subscriber is not None:
#             return await run_callback(subscriber, msg)
#         return fail(err=MissingSubscriber(debug={"subject": topic}))

#     @t.override
#     def register(self, *services: Service) -> None:
#         for service in services:
#             for subscriber in service.subscribers:
#                 if subscriber.with_reply:
#                     self.handlers[subscriber.topic] = subscriber
#                 else:
#                     self.subscribers.append(subscriber)

#     @t.override
#     async def connect(self):
#         while not self.stop_signal.is_set() or not self.queue.empty():
#             try:
#                 msg = await self.queue.get()
#                 tasks = [
#                     create_task(run_callback(subscriber, msg))
#                     for subscriber in self.subscribers
#                     if is_subscribed(subscriber, msg)
#                 ]
#                 await wait(tasks)
#             except Exception as exc:
#                 print(exc)
